{    StarsXY 1.0 -     map-   Stars!    }
{    Tested on 2.7i                 -     D4   }

{   ========  Sorry : Comments available in Russian only =========   }

{    by Alexander Kramarenko (kram@beep.ru), Smolensk, Russia, 1999  }
{    my commercial site is : www.aklabs.com                          }

{    Special thanks to Konstantin Lanin <klanin@tavrich.ru>          }

// --------------------------------------------------------------
//  ,     :
//   TStarPack
//      -    id- 
//          x,y -  
//          id  -  

//    TStarsXY :
//   constructor Create(fn: String);
//      -   TStarsXY   xy- fn
//
//   destructor Free;
//      -   TStarsXY
//
//   procedure PackAndWrite(ofn : String);
//      -   .
//
//   function Boundaries:TRect;
//      -     
//
//   procedure SetStar(n,x,y : Integer; id : Integer = -1);
//      -    n.
//          x,y -  
//          id -   , -1 -  .
//
//   function GetStar(n : Integer):TStarPack;
//       -      n
//
//   property PlayersCount:word read nPlayers;
//       -     
//
//   property StarsCount:word read nStars;
//       -     
// --------------------------------------------------------------

unit StarsXY;

interface
uses Windows, SysUtils;

ResourceString
 erFileNotOpen = 'XY-file not open';
 erPackError   = 'Can not pack information on planets';
 erStarNotFound = 'Specified star does not exist';
 erOverTheUniverse = 'Specified coordinates of star are beyond the scope of seen part universe';
 erStarExists = 'Such star already exists';
 erUnableLoad = 'The game file appears to be corrupt';

const
 usTiny = 0;
 usSmall = 1;
 usMedium = 2;
 usLarge = 3;
 usHuge = 4;

 usSizes : Array[0..4] of String = ('Tiny', 'Small', 'Medium', 'Large', 'Huge');

type
  THdr = packed record
   cb : word; // Size of the record in bytes not counting this header - 10bit.
   rt : byte; // Record Type (rtBOF etc) - 6 bit.
   end;

  TRTBOF = packed record
   rgid : String[4];

   lidGame : dWord;  // Game ID

   verInc,           // 5 - verInc
   verMinor,         // 7 - verMinor
   verMajor : Word;  // 4 - verMajor
   turn     : Word;

   iPlayer,          // 5 - iPlayer
   lSaltTime : Word; // 11 - lSaltTime

   dt,              // 8 - File type dtXY, dtHost, dtLog, dtHist etc.
   fDone,           // 1 - Player has submitted this turn (dtLog only).
   fInUse,          // 1 - Host instance is using this file (dtHost, dtTurn).
   fMulti,          // 1 - Multiple turns in this file (dtTurn only).
   fGameOver,       // 1 - A winner has been declared
   fShareware,      // 1 - The shareware version
   unused : Word;   // 3 - Unused
  end;

  TRTEOF = word;

  TStarPack = packed record
    x,
    y,
    id : dWord;
   end;


  TStarsXY = class
    protected
       star_table : array of TStarPack;
       bin_image : String;
       bin_length : Integer;
       _valid : Boolean;
       MinDist : Word;

       // pointers inside file
       bof : TRTBOF;
       eof : TRTEOF;
       lenPack, posPack : Word;

       // external interface stars positions
       nstars : Word;
       xmin, ymin, xmax, ymax  : dWord;
       nplayers : word;

       function GetHeader(dat:Pointer):THdr;
       function GetBof(dat:Pointer):TRTBOF;
       function GetStarPack(dat:Pointer):TStarPack;
       function SetStarPack(dat:TStarPack):dWord;
     public
       constructor Create(fn: String);
       destructor Free;
       procedure PackAndWrite(ofn : String; sort:Boolean = true);
       function Boundaries:TRect;
       procedure SetStar(n,x,y : Integer; id : Integer = -1; Check : boolean = true);
       function GetStar(n : Integer):TStarPack;

       property PlayersCount:word read nPlayers;
       property StarsCount:word read nStars;
       property Opened:boolean read _valid;
       property MinDistance:Word read MinDist write MinDist;
       function GetUniverseSize:byte;
       procedure SetUniverseSize(sz:Byte);
   end;

  function GetBits(d:dWord; from, len:Byte):dWord;
  procedure SetBits(var d:dWord; from, len:Byte; data:dWord);
  function GetBitsIm(dat:Pointer; from, len : byte;  sz:byte = 2):dWord;

implementation

const
 dtXY   = 0;
 dtLog  = 1;
 dtHost = 2;
 dtTurn = 3;
 dtHist = 4;
 dtRace = 5;
 dtMax  = 6;

 rtEOF  = 0;  // End Of File Record Type
 rtGAME = 7;  // GAME Record Type
 rtBOF  = 8;  // Begining Of File Record Type

 szMagic = 'J3J3';
 SizeRTBOF = 16;


function GetBits(d:dWord; from, len:Byte):dWord;
var
  max : byte;
begin
     max := len + from - 1;
     Result := dWord(dWord(d shl (31-max))   shr (from+(31-max)))
end;

procedure SetBits(var d:dWord; from, len:Byte; data:dWord);
var
  fl : dWord;
begin
     fl := not dWord(GetBits($FFFFFFFF, from, len) shl from);
     d := (d and fl) or (dWord(GetBits(Data, 0, len) shl from));
end;

function GetBitsIm(dat:Pointer; from, len : byte;  sz:byte = 2):dWord;
var dw : dword;
begin
     move(dat^, dw, sz);
     Result := GetBits(dw, from, len);
end;



{ TStarsXY }

function TStarsXY.GetHeader(dat:Pointer):THdr;
var w : word;
begin
     Move(dat^, w, 2);
     Result.cb := GetBits(w, 0, 10);
     Result.rt := GetBits(w, 10, 6);
end;

function TStarsXY.GetBof(dat:Pointer):TRTBOF;
var
  bin_img : Array[1..SizeRTBOF] of char;
begin
  move(dat^, bin_img[1], SizeRTBOF);
  with Result do
   begin
    rgid[0] := #4; Move(bin_img[1], rgid[1], 4);
    Move(bin_img[5], lIdGame, 4);
    verInc   := GetBitsIm(@bin_img[9], 0, 5);
    verMinor := GetBitsIm(@bin_img[9], 5, 7);
    verMajor := GetBitsIm(@bin_img[9], 12, 4);
    Move(bin_img[11], turn, 2);

    iPlayer  := GetBitsIm(@bin_img[13], 0, 5);
    lSaltTime  := GetBitsIm(@bin_img[13], 5, 11);

    dt := GetBitsIm(@bin_img[15], 0, 8);
    fDone := GetBitsIm(@bin_img[15], 8, 1);
    fInUse := GetBitsIm(@bin_img[15], 9, 1);
    fMulti := GetBitsIm(@bin_img[15], 10, 1);
    fGameOver := GetBitsIm(@bin_img[15], 11, 1);
    fShareware:= GetBitsIm(@bin_img[15], 12, 1);
    unused := GetBitsIm(@bin_img[15], 13, 3);
   end;
end;

function TStarsXY.GetStarPack(dat:Pointer):TStarPack;
var
  bin_img : dWord;
begin
  move(dat^, bin_img, 4);
  with Result do
   begin
    x := GetBitsIm(@bin_img, 0, 10, 4);
    y := GetBitsIm(@bin_Img, 10, 12, 4);
    id := GetBitsIm(@bin_img, 22, 10, 4);
   end;
end;

function TStarsXY.SetStarPack(dat:TStarPack):dWord;
begin
  with dat do
   begin
    Result := 0;
    SetBits(Result, 0, 10, x);
    SetBits(Result, 10, 12, y);
    SetBits(Result, 22, 10, id);
   end;
end;

constructor TStarsXY.Create(fn: String);
var
   f : file;
   _x, _y  : dWord;
   Pos, i : Integer;
   sp : TStarPack;
   hdr : THdr;
begin
  MinDist := 5;

  AssignFile(f, fn);
  try
   Reset(f,1);
   bin_length := FileSize(f);
   SetLength(bin_image, bin_length);
   BlockRead(f, bin_image[1], bin_length);
  finally
   CloseFile(f);
  end;

  // check the validity of the file
  hdr := GetHeader(@bin_image[1]);
  if (hdr.rt <> rtBOF) or (hdr.cb < SizeRTBOF) then
   raise Exception.Create(erUnableLoad);

  bof := GetBof(@bin_image[3]);

  if (szMagic <> bof.rgid) or (bof.dt <> dtXY) then
   raise Exception.Create(erUnableLoad);

  // skip game record
  pos := 3 + SizeRTBOF;
  hdr := GetHeader(@bin_image[pos]); inc(pos,2+hdr.cb);
  PosPack := pos; lenPack := hdr.cb;

  // now jump to eof
  pos := bin_length - 4 + 1;
  hdr := GetHeader(@bin_image[pos]); inc(pos,hdr.cb);
  if hdr.rt <> rtEOF then
   raise Exception.Create(erUnableLoad);

  move(bin_image[pos], nplayers, 2);

// file looks valid
  _valid := true;

// unpack starpack stuff
  nstars := Trunc(((pos-2) - PosPack) / 4);

  ymin := 10000;
  ymax := 0;
  _x := 1000;

  SetLength(Star_table, nstars);
  for i := 0 to nstars-1 do
   begin
    sp := GetStarPack(@bin_image[posPack + i*4]);

    _x := _x + sp.x;
    _y := sp.y;

    if _y < ymin then ymin := _y;
    if _y > ymax then ymax := _y;

    Star_table[i].x := _x;
    Star_table[i].y := _y;
    Star_table[i].id := sp.id;
   end;

  xmin := 1000;
  xmax := _x + 1;
end;

procedure TStarsXY.PackAndWrite(ofn : String; Sort:Boolean = true);
var
  i, j, x, dx : Integer;
  sp : TStarPack;
  dw : dWord;
  f : file;
begin
  if not _valid then
   raise Exception.Create(erFileNotOpen);


  // sort the array along x/y
  if Sort then
   for i := 0 to nstars-2 do
    for j := i+1 to nstars-1 do
     if (star_table[j].x < star_table[i].x) or
         ((star_table[j].x = star_table[i].x) and
 	 (star_table[j].y < star_table[i].y) ) then
      begin
       // swap the elements
       sp := star_table[j];
       star_table[j] := star_table[i];
       star_table[i] := sp;
      end;

  // now pack the data
  x := 1000;
  for i := 0 to nstars-1 do
   begin
    sp := star_table[i];
    sp.x := sp.x - x;
    if (sp.x > 1023) then
     raise Exception.Create(erPackError);
    x := star_table[i].x;
    dw := SetStarPack(sp);
    move(dw, bin_image[posPack + i*4], 4);
  end;

  // write out the file
  AssignFile(f, ofn);
  try
   Rewrite(f,1);
   BlockWrite(f, bin_image[1], bin_length);
  finally
   CloseFile(f);
  end;
end;

function TStarsXY.Boundaries:TRect;
begin
  if not _valid then
   raise Exception.Create(erFileNotOpen);

  with Result do
   begin
    Left := xmin;
    Right := xmax;
    Top := ymin;
    Bottom := ymax;
   end;
end;

procedure TStarsXY.SetStar(n,x,y : Integer; id : Integer = -1; Check : boolean = true);
var rnd, i, _x, _y : Integer;
    MinD : Integer;
begin
  if not _valid then
   raise Exception.Create(erFileNotOpen);

  _x := x;  _y := y;
  if Check then
   begin
    _x := x + Round(Random(7)-3);
    _y := y + Round(Random(7)-3);
   end;

  if (n < 0) or (n >= nstars) then
   raise Exception.Create(erStarNotFound);

  if (_x < xmin) or (_x > xmax) or (_y < ymin) or (_y > ymax) then
   raise Exception.Create(erOverTheUniverse);

  MinD := MinDist;

  // check for superposition
  if Check then
   for i := 0 to nstars-1 do
    if (i <> n) and
       (star_table[i].x > _x-MinD) and (star_table[i].x < _x+MinD) and
       (star_table[i].y > _y-MinD) and (star_table[i].y < _y+MinD) then
     raise Exception.Create(erStarExists);

  star_table[n].x := _x;
  star_table[n].y := _y;

  if id <> -1 then
   star_table[n].id := id;
end;


function TStarsXY.GetStar(n : Integer):TStarPack;
begin
  if not _valid then
   raise Exception.Create(erFileNotOpen);

  if (n < 0) or (n >= nstars) then
   raise Exception.Create(erStarNotFound);

  Result := star_table[n];
end;


destructor TStarsXY.Free;
begin
     {}
end;

function TStarsXY.GetUniverseSize: byte;
var
   r:TRect;
   sz : Integer;
begin
     if not _valid then
      raise Exception.Create(erFileNotOpen);
     r := Boundaries;
     sz := r.Right-r.Left;
     Result := usHuge;
     if sz <= 1600 then Result:= usLarge;
     if sz <= 1200 then Result := usMedium;
     if sz <= 800 then Result := usSmall;
     if sz <= 400 then Result := usTiny;
end;

procedure TStarsXY.SetUniverseSize(sz: Byte);
var
  sm : Integer;
begin
     sm := (sz+1)*400-2;
     xmin := 1000;
     ymin := 1000;

     xmax := sm + xmin;
     ymax := sm + ymin;
end;

end.
